#!/usr/bin/python3
from __future__ import division
import os
import sys
import numpy as np

class ploadparticles(object):
	def __init__(self, ns, w_dir=None, datatype=None, ptype=None, chnum=None):
		"""
		Loads the Particle datafile.

		**Inputs**:
		ns 	-- Step Number of the data file\n
		w_dir 	 -- path to the directory which has the data files\n
		datatype -- Datatype (default is set to read .dbl data files)
		ptype 	 -- A string denoting the type of particles ('LP', 'CR', 'DUST' etc. Default is 'CR')
		chnum    -- 2 digit integer denoting chunk number
		(Only used if ptype = 'LP' to read particles.nnnn_chxx.dbl file, where nnnn is 4 digit integer denotong ns and xx is a 2 digit integer for chunk number : Default value is 0)
		
		**Outputs**:
		pyPLUTO pload object whose keys are arrays of data values.
		"""
		self.Nstep = ns
		if w_dir is None:
			w_dir = os.getcwd() + '/'
		self.wdir = w_dir
		if datatype is None:
			datatype = "dbl"
		self.datatype = datatype
		if ptype == 'LP' and self.datatype in ['dbl', 'flt']:
			if chnum is None: chnum = 0 #by default it reads first file. 
			self.fname = self.wdir+"particles.%04d_ch%02d.%s"%(ns, chnum, self.datatype)
		else:
			self.fname = self.wdir+"particles.%04d.%s"%(ns, self.datatype)
		if self.datatype == 'vtk':
			Part_dictionary = self.ReadVTKParticleFile()
		else:
			Part_dictionary = self.ReadBinParticleFile()
		for keys in Part_dictionary:
			object.__setattr__(self, keys, Part_dictionary.get(keys))

	def ReadVTKParticleFile(self):
		print("Reading particle file : %s"%self.fname)
		fp = open(self.fname,'rb')
		nfields = 0
	
		while True:
			line = fp.readline()
			try:
				line.split()[0]
			except IndexError as error:
				pass
			else:
				if line.split()[0] == b'POINTS':
					nparts = int(line.decode().split()[1])
					dtype_tup = str(nparts*3)+'>f4'
					nb = np.dtype(dtype_tup).itemsize
					vtkvar_buf = np.frombuffer(fp.read(nb), dtype=np.dtype(dtype_tup))
					coords = vtkvar_buf.reshape(nparts,3)
					val_dict = {'Totparticles':nparts, 'x1':coords[:,0],'x2':coords[:,1],'x3':coords[:,2]}
					nfields += 3
				
				if line.split()[0] == b'SCALARS':
					vars = line.decode().split()[1]
				if line.split()[0] == b'LOOKUP_TABLE':
					if vars == 'Identity': 
						dtype_tup = str(nparts)+'>i4'
						field_name = 'id'
					elif vars == 'tinj': 
						dtype_tup = str(nparts)+'>f4'
						field_name = 'tinj'
					elif vars == 'Color': 
						dtype_tup == str(nparts)+'>f4'
						field_name = 'color'
					nb = np.dtype(dtype_tup).itemsize
					vtkvar_buf = np.frombuffer(fp.read(nb), dtype=np.dtype(dtype_tup))
					val_dict.update({field_name:vtkvar_buf})
					nfields += 1
				if line.split()[0] == b'VECTORS':
					vars = line.decode().split()[1]
					dtype_tup = str(nparts*3)+'>f4'
					nb = np.dtype(dtype_tup).itemsize
					vtkvar_buf = np.frombuffer(fp.read(nb), dtype=np.dtype(dtype_tup))
					vels = vtkvar_buf.reshape(nparts,3)
					val_dict.update({'vx1':vels[:,0], 'vx2':vels[:,1], 'vx3':vels[:,2]})
					nfields += 3
				else:
					pass
			if line == b'':
				break
		val_dict.update({'nfields':nfields})
		return val_dict

	def ReadBinParticleFile(self):
		print("Reading particle file : %s"%self.fname)
		fp = open(self.fname, "rb")
		val_dict = {}
		h_lines = 0

		#READ HEADER. 
		with open(self.fname,"rb") as f:
			for line in f:
				if line.startswith(b'#'):
					if line.split()[1] != b'PLUTO': 
						val_dict.update({line.split()[1].decode('utf8'):[i.decode('utf8') for i in line.split()[2:]]})
					h_lines += 1
		
		#SKIP HEADER LINES	
		cnt = 0
		while (cnt < h_lines):
			fp.readline()
			cnt += 1
        
		#READ DATA
		data_ = fp.read()
		fp.close()

		#SORT DATA INTO DICTIONARY BASED ON DATATYPE.
		if self.datatype == 'flt':
			dt = np.dtype({'names':val_dict['field_names'], 'formats':['('+i+',)<f' for i in val_dict['field_dim']]})
		else:
			dt = np.dtype({'names':val_dict['field_names'], 'formats':['('+i+',)<d' for i in val_dict['field_dim']]})

		val_ = np.frombuffer(data_,dtype=dt)

		for i in range(len(val_dict['field_names'])):
			name = val_dict['field_names'][i]
			if int(val_dict['field_dim'][i]) == 1:
				val_dict.update({name:val_[name].flatten()})
			else:
				val_dict.update({name:val_[name]})

		#OUTPUT IS A DICTIONARY. 
		return val_dict



		




def read_time(w_dir=None,datatype=None):
    """ read the dbl.out file to get the time information for each dump

    **Inputs**:
    
      w_dir -- path to the directory which has the dbl.out(or flt.out) and the data\n
      datatype -- If the data is of 'float' type then datatype = 'flt' else by default the datatype is set to 'dbl' (Double precision).

        **Outputs**:
    
      This function returns a dictionary with following keywords - \n

      nlast -- The ns for the last file saved.\n
      time -- The simulation time for the last file saved.\n
      dt -- The time step dt for the last file. \n
      Nstep -- The Nstep value for the last file saved.


    **Usage**:
    
      In case the data is 'float'.

      ``wdir = /path/to/data/directory``\n
      ``import pyPLUTO as pp``\n
      ``A = pp.nlast_info(w_dir=wdir,datatype='float')``    
    """
    if w_dir is None: w_dir=os.getcwd()+'/'
    if datatype == 'flt':
        fname_v = w_dir+"flt.out"
    elif datatype == 'vtk':
        fname_v = w_dir+"vtk.out"
    else:
        fname_v = w_dir+"dbl.out"

    time = []
    file = open(fname_v, "r")
    for line in file:
      fields = line.split()
      time.append(fields[1])
            
    return time
    


def nlast_info(w_dir=None,datatype=None):
    """ Prints the information of the last step of the simulation as obtained from out files

    **Inputs**:
    
      w_dir -- path to the directory which has the dbl.out(or flt.out) and the data\n
      datatype -- If the data is of 'float' type then datatype = 'flt' else by default the datatype is set to 'dbl' (Double precision).

        **Outputs**:
    
      This function returns a dictionary with following keywords - \n

      nlast -- The ns for the last file saved.\n
      time -- The simulation time for the last file saved.\n
      dt -- The time step dt for the last file. \n
      Nstep -- The Nstep value for the last file saved.


    **Usage**:
    
      In case the data is 'float'.

      ``wdir = /path/to/data/directory``\n
      ``import pyPLUTO as pp``\n
      ``A = pp.nlast_info(w_dir=wdir,datatype='float')``    
    """
    if w_dir is None: w_dir=os.getcwd()+'/'
    if datatype == 'flt':
        fname_v = w_dir+"flt.out"
    elif datatype == 'vtk':
        fname_v = w_dir+"vtk.out"
    else:
        fname_v = w_dir+"dbl.out"

    with open(fname_v, 'r') as f:
        lines = f.read().splitlines()
        last_line = lines[-1].split()
        
    nlast = int(last_line[0])
    SimTime =  float(last_line[1])
    Dt = float(last_line[2])
    Nstep = int(last_line[3])
        
    print("------------TIME INFORMATION--------------")
    print('nlast = %d'%nlast)
    print('time  =%f'%SimTime)
    print('dt    =%f'%Dt)
    print('Nstep =%d'%Nstep)
    print("-------------------------------------------")
        
    return {'nlast':nlast,'time':SimTime,'dt':Dt,'Nstep':Nstep}
    



#!/usr/bin/python3
from __future__ import division
import os
import sys
import numpy as np

class ploadparticles(object):
	def __init__(self, ns, w_dir=None, datatype=None, ptype=None, chnum=None):
		"""
		Loads the Particle datafile.

		**Inputs**:
		ns 			 -- Step Number of the data file\n
		w_dir 	 -- path to the directory which has the data files\n
		datatype -- Datatype (default is set to read .dbl data files)
		ptype 	 -- A string denoting the type of particles ('LP', 'CR', 'DUST' etc. Default is 'CR')
		chnum    -- 2 digit integer denoting chunk number
		(Only used if ptype = 'LP' to read particles.nnnn_chxx.dbl file, where nnnn is 4 digit integer denotong ns and xx is a 2 digit integer for chunk number : Default value is 0)
		
		**Outputs**:
		pyPLUTO pload object whose keys are arrays of data values.

		"""
		self.Nstep = ns
		
		
		if w_dir is None:
			w_dir = os.getcwd() + '/'
		
		self.wdir = w_dir

		if datatype is None:
			datatype = "dbl"
		self.datatype = datatype
		if (self.datatype == "vtk"):
			print("Particle VTK File is not yet supported. Choose either dbl or flt")
			sys.exit()
		
		if ptype == 'LP':
			if chnum is None : chnum=0
			self.fname = self.wdir+"particles.%04d_ch%02d.%s"%(ns, chnum, self.datatype)
		else:
			self.fname = self.wdir+"particles.%04d.%s"%(ns, self.datatype)
			
		
		Part_dictionary = self.ReadParticleFile()
		for keys in Part_dictionary:
			object.__setattr__(self, keys, Part_dictionary.get(keys))

	def ReadParticleFile(self):
		print("Reading particle file : %s"%self.fname)
		fp = open(self.fname,'rb')
		#GET THE HEADER ROUTINE HERE
		h_lines = 0
		header = []
		while (True):
			ln = fp.readline()
			lns = ln.split()
			if lns[0] == b'#':
				if lns[1] != b'PLUTO': header.append((lns[1].decode('utf8'), [i.decode('utf8') for i in lns[2:]]))
			else:
				break
			h_lines += 1
		self.Head_Dict = dict(header)
		fp.close()
		
		# SKIP HEADER LINES
		cnt = 0
		fp = open(self.fname,'rb')
		while (cnt < h_lines):
			fp.readline()
			cnt += 1

		data_ = fp.read()
		#Create Data type tuple
		
		data_typ_tup = []
		for i in range(int(self.Head_Dict['nfields'][0])):
			if self.datatype == 'flt':
				data_typ_tup.append((self.Head_Dict['field_names'][i], np.float32, int(self.Head_Dict['field_dim'][i])))
			else:
				data_typ_tup.append((self.Head_Dict['field_names'][i], np.double, int(self.Head_Dict['field_dim'][i])))
	
		dt = np.dtype(data_typ_tup)
		
		val_ = np.frombuffer(data_,dtype=dt)
		val_dict = {name:val_[name] for name in self.Head_Dict['field_names']}
		return val_dict




# -*- coding: utf-8 -*-
import os
from . import pload
from . import ploadparticles
from . import Tools
from . import Image


def nlast_info(w_dir=None,datatype=None):
	""" 
	Prints the information of the last step of the simulation as obtained from out files

	**Inputs**:
	
	  w_dir -- path to the directory which has the dbl.out(or flt.out) and the data\n
	  datatype -- If the data is of 'float' type then datatype = 'flt' else by default the datatype is set to 'dbl' (Double precision).

    **Outputs**:
	
	  This function returns a dictionary with following keywords - \n

	  nlast -- The ns for the last file saved.\n
	  time -- The simulation time for the last file saved.\n
	  dt -- The time step dt for the last file. \n
	  Nstep -- The Nstep value for the last file saved.


	**Usage**:
	
	  In case the data is 'float'.

	  ``wdir = /path/to/data/directory``\n
	  ``import pyPLUTO as pp``\n
	  ``A = pp.nlast_info(w_dir=wdir,datatype='float')``	
	"""

	if w_dir is None: w_dir = os.getcwd()+'/'
	if datatype == 'flt':
		fname_v = w_dir+"flt.out"
	elif datatype == 'vtk':
		fname_v = w_dir+"vtk.out"
	else:
		fname_v = w_dir+"dbl.out"

	with open(fname_v, 'r') as f:
		lines = f.read().splitlines()
		last_line = lines[-1].split()
		
	nlast = int(last_line[0])
	SimTime =  float(last_line[1])
	Dt = float(last_line[2])
	Nstep = int(last_line[3])
	    
	print("------------TIME INFORMATION--------------")
	print('nlast = %d'%nlast)
	print('time  =%f'%SimTime)
	print('dt    =%f'%Dt)
	print('Nstep =%d'%Nstep)
	print("-------------------------------------------")
	    
	return {'nlast':nlast,'time':SimTime,'dt':Dt,'Nstep':Nstep}
    




#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name="pysunrunner",
    version="4.4.7",
    packages=find_packages(),
    install_requires=[
        'numpy==2.0.0',
        'matplotlib==3.9.1',
        'scipy==1.14.0',
        'pillow==10.4.0',
    ],
    author="Jackson Riley (pyPLUTO routines by Dr. Bhargav Vaidya)",
    author_email="jriley@predsci.com",
    description="A package for imaging and plotting data from sunrunner simulations",
    url="https://github.com/BadOmen59/PySunRunner", 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)




from __future__ import division
import os
import numpy as np
from matplotlib.pyplot import *
from matplotlib.mlab import *
from . import Tools

class Image(object):
	def pldisplay(self, D, var,**kwargs):
		""" This method allows the user to display a 2D data using the
			 matplotlib's pcolormesh.

		**Inputs:**
			D   -- pyPLUTO pload object.\n
			var -- 2D array that needs to be displayed.
		*Required Keywords:*
			x1 -- The 'x' array\n
			x2 -- The 'y' array
		*Optional Keywords:*
			vmin -- The minimum value of the 2D array (Default : min(var))\n
			vmax -- The maximum value of the 2D array (Default : max(var))\n
			title -- Sets the title of the image.\n
			label1 -- Sets the X Label (Default: 'XLabel')\n
			label2 -- Sets the Y Label (Default: 'YLabel')\n
			polar -- A list to project Polar data on Cartesian Grid.\n
				polar = [True, True] -- Projects r-phi plane.\n
				polar = [True, False] -- Project r-theta plane.\n
				polar = [False, False] -- No polar plot (Default)\n
			cbar -- Its a tuple to set the colorbar on or off. \n
				cbar = (True,'vertical') -- Displays a vertical colorbar\n
				cbar = (True,'horizontal') -- Displays a horizontal colorbar\n
				cbar = (False,'') -- Displays no colorbar.
		**Usage:**
			``import pyPLUTO as pp``\n
			``wdir = '/path/to/the data files/'``\n
			``D = pp.pload(1,w_dir=wdir)``\n
			``I = pp.Image()``\n
			``I.pldisplay(D, D.v2, x1=D.x1, x2=D.x2, cbar=(True,'vertical'),\
			title='Velocity',label1='Radius',label2='Height')``
		"""
		
		x1 = kwargs.get('x1')
		x2 = kwargs.get('x2')
		var = var.T

		f1 = figure(kwargs.get('fignum',1), figsize=kwargs.get('figsize',[10,10]),
									 dpi=80, facecolor='w', edgecolor='k')
		ax1 = f1.add_subplot(111)
		ax1.set_aspect('equal')

		if kwargs.get('polar',[False,False])[0]:
			xx, yy = self.getPolarData(D,kwargs.get('x2'),rphi=kwargs.get('polar')[1])
			pcolormesh(xx,yy,var,vmin=kwargs.get('vmin',np.min(var)),vmax=kwargs.get('vmax',np.max(var)))
		else:
			ax1.axis([np.min(x1),np.max(x1),np.min(x2),np.max(x2)])
			pcolormesh(x1,x2,var,vmin=kwargs.get('vmin',np.min(var)),vmax=kwargs.get('vmax',np.max(var)))
			 
		title(kwargs.get('title',"Title"),size=kwargs.get('size'))
		xlabel(kwargs.get('label1',"Xlabel"),size=kwargs.get('size'))
		ylabel(kwargs.get('label2',"Ylabel"),size=kwargs.get('size'))
		if kwargs.get('cbar',(False,''))[0] == True:
			colorbar(orientation=kwargs.get('cbar')[1])
	 
	 
	def multi_disp(self,*args,**kwargs):
		mvar = []
		for arg in args:
			mvar.append(arg.T)
			 
		xmin = np.min(kwargs.get('x1'))
		xmax = np.max(kwargs.get('x1'))
		ymin = np.min(kwargs.get('x2'))
		ymax = np.max(kwargs.get('x2'))
		mfig = figure(kwargs.get('fignum',1),figsize=kwargs.get('figsize',[10,10]))
		Ncols = kwargs.get('Ncols')
		Nrows = len(args)//Ncols
		mprod = Nrows*Ncols
		dictcbar=kwargs.get('cbar',(False,'','each'))

		for j in range(mprod):
			mfig.add_subplot(Nrows,Ncols,j+1)
			pcolormesh(kwargs.get('x1'),kwargs.get('x2'), mvar[j])
			axis([xmin,xmax,ymin,ymax])
			gca().set_aspect('equal')
							 
			xlabel(kwargs.get('label1',mprod*['Xlabel'])[j])
			ylabel(kwargs.get('label2',mprod*['Ylabel'])[j])
			title(kwargs.get('title',mprod*['Title'])[j])
			if (dictcbar[0] == True) and (dictcbar[2] =='each'):
				colorbar(orientation=kwargs.get('cbar')[1])
			if dictcbar[0] == True and dictcbar[2]=='last':
				if (j == np.max(range(mprod))):colorbar(orientation=kwargs.get('cbar')[1])
				
	def oplotbox(self, AMRLevel, lrange=[0,0], cval=['b','r','g','m','w','k'],\
										islice=-1, jslice=-1, kslice=-1,geom='CARTESIAN'):
		"""
		This method overplots the AMR boxes up to the specified level.
		**Input:**
			AMRLevel -- AMR object loaded during the reading and stored in the pload object
		*Optional Keywords:*
			lrange     -- [level_min,level_max] to be overplotted. By default it shows all the loaded levels\n
			cval       -- list of colors for the levels to be overplotted.\n
			[ijk]slice -- Index of the 2D slice to look for so that the adequate box limits are plotted.
										By default oplotbox considers you are plotting a 2D slice of the z=min(x3) plane.\n
			geom       -- Specified the geometry. Currently, CARTESIAN (default) and POLAR geometries are handled.
		"""

		nlev = len(AMRLevel)
		lrange[1] = min(lrange[1],nlev-1)
		npl  = lrange[1]-lrange[0]+1
		lpls = [lrange[0]+v for v in range(npl)]
		cols = cval[0:nlev]
		# Get the offset and the type of slice
		Slice = 0 ; inds = 'k'
		xx = 'x' ; yy ='y'
		if (islice >= 0):
				Slice = islice + AMRLevel[0]['ibeg'] ; inds = 'i'
				xx = 'y' ; yy ='z'
		if (jslice >= 0):
				Slice = jslice + AMRLevel[0]['jbeg'] ; inds = 'j'
				xx = 'x' ; yy ='z'
		if (kslice >= 0):
				Slice = kslice + AMRLevel[0]['kbeg'] ; inds = 'k'
				xx = 'x' ; yy ='y'
					 
		# Overplot the boxes
		for il in lpls:
			level = AMRLevel[il]
			for ib in range(level['nbox']):
				box = level['box'][ib]
				if ((Slice-box[inds+'b'])*(box[inds+'e']-Slice) >= 0):
					if (geom == 'CARTESIAN'):
						x0 = box[xx+'0'] ; x1 = box[xx+'1']
						y0 = box[yy+'0'] ; y1 = box[yy+'1']
						plot([x0,x1,x1,x0,x0],[y0,y0,y1,y1,y0],color=cols[il])
					elif (geom == 'POLAR') or (geom == 'SPHERICAL'):
						dn = np.pi/50.
						x0 = box[xx+'0'] ; x1 = box[xx+'1']
						y0 = box[yy+'0'] ; y1 = box[yy+'1']
						if y0 == y1:
							y1 = 2*np.pi+y0 - 1.e-3
						xb = np.concatenate([
										[x0*np.cos(y0),x1*np.cos(y0)],\
										x1*np.cos(np.linspace(y0,y1,num=int(abs(y0-y1)/dn) )),\
										[x1*np.cos(y1),x0*np.cos(y1)],\
										x0*np.cos(np.linspace(y1,y0,num=int(abs(y0-y1)/dn)))])
						yb = np.concatenate([
										[x0*np.sin(y0),x1*np.sin(y0)],\
										x1*np.sin(np.linspace(y0,y1,num=int(abs(y0-y1)/dn))),\
										[x1*np.sin(y1),x0*np.sin(y1)],\
										x0*np.sin(np.linspace(y1,y0,num=int(abs(y0-y1)/dn)))])
						plot(xb,yb,color=cols[il])
				
	def field_interp(self,var1,var2,x,y,dx,dy,xp,yp):
		""" 
			This method interpolates value of vector fields (var1 and var2) on field points (xp and yp).
			The field points are obtained from the method field_line.
			**Inputs:**
	 			var1 -- 2D Vector field in X direction\n
				var2 -- 2D Vector field in Y direction\n
				x -- 1D X array\n
				y -- 1D Y array\n
				dx -- 1D grid spacing array in X direction\n
				dy -- 1D grid spacing array in Y direction\n
				xp -- field point in X direction\n
				yp -- field point in Y direction\n	 
			**Outputs:**
				A list with 2 elements where the first element corresponds to the interpolate field
				point in 'x' direction and the second element is the field point in 'y' direction.
		"""
		q=[]
		U = var1
		V = var2
		i0 = np.abs(xp-x).argmin()
		j0 = np.abs(yp-y).argmin()
		scrhUx = np.interp(xp,x,U[:,j0])
		scrhUy = np.interp(yp,y,U[i0,:])
		q.append(scrhUx + scrhUy - U[i0,j0])
		scrhVx = np.interp(xp,x,V[:,j0])
		scrhVy = np.interp(yp,y,V[i0,:])
		q.append(scrhVx + scrhVy - V[i0,j0])
		return q
	 
	def field_line(self,var1,var2,x,y,dx,dy,x0,y0):
		""" This method is used to obtain field lines (same as fieldline.pro in PLUTO IDL tools).
		**Inputs:**
			var1 -- 2D Vector field in X direction\n
			var2 -- 2D Vector field in Y direction\n
			x -- 1D X array\n
			y -- 1D Y array\n
			dx -- 1D grid spacing array in X direction\n
			dy -- 1D grid spacing array in Y direction\n
			x0 -- foot point of the field line in X direction\n
			y0 -- foot point of the field line in Y direction\n
		**Outputs:**
			This routine returns a dictionary with keys - \n
			qx -- list of the field points along the 'x' direction.
			qy -- list of the field points along the 'y' direction.
		**Usage:**
			See the myfieldlines routine for the same.
		"""
		xbeg = x[0] - 0.5*dx[0]
		xend = x[-1] + 0.5*dx[-1]
		
		ybeg = y[0]  - 0.5*dy[0]
		yend = y[-1] + 0.5*dy[-1]

		inside_domain = x0 > xbeg and x0 < xend and y0 > ybeg and y0 < yend

		MAX_STEPS = 5000
		xln_fwd = [x0]
		yln_fwd = [y0]
		xln_bck = [x0]
		yln_bck = [y0]
		rhs = []
		k = 0
	 
		while inside_domain == True:
			R1 = self.field_interp(var1,var2,x,y,dx,dy,xln_fwd[k],yln_fwd[k])
			dl = 0.5*np.max(np.concatenate((dx,dy)))/(np.sqrt(R1[0]*R1[0] + R1[1]*R1[1] + 1.e-14))
			xscrh = xln_fwd[k] + 0.5*dl*R1[0]
			yscrh = yln_fwd[k] + 0.5*dl*R1[1]
			
			R2 = self.field_interp(var1,var2,x,y,dx,dy,xln_fwd[k],yln_fwd[k])
			
			xln_one = xln_fwd[k] + dl*R2[0]
			yln_one = yln_fwd[k] + dl*R2[1]
			
			xln_fwd.append(xln_one)
			yln_fwd.append(yln_one)
			inside_domain = xln_one > xbeg and xln_one < xend and yln_one > ybeg and yln_one < yend
			inside_domain = inside_domain and (k < MAX_STEPS-3)
			k = k + 1
	
		k_fwd = k
		qx = np.asarray(xln_fwd[0:k_fwd])
		qy = np.asarray(yln_fwd[0:k_fwd])
		flines={'qx':qx,'qy':qy}
		
		return flines
		 
		 
	def myfieldlines(self,Data,x0arr,y0arr,stream=False,**kwargs):
		""" This method overplots the magnetic field lines at the footpoints given by (x0arr[i],y0arr[i]).
		**Inputs:**
			Data -- pyPLUTO.pload object\n
			x0arr -- array of x co-ordinates of the footpoints\n
			y0arr -- array of y co-ordinates of the footpoints\n
			stream -- keyword for two different ways of calculating the field lines.\n
			True -- plots contours of rAphi (needs to store vector potential)\n
			False -- plots the fieldlines obtained from the field_line routine. (Default option)
		*Optional Keywords:*
			colors -- A list of matplotlib colors to represent the lines. The length of this list should be same as that of x0arr.\n
			lw -- Integer value that determines the linewidth of each line.\n
			ls -- Determines the linestyle of each line.
		**Usage:**
			Assume that the magnetic field is a given as **B** = B0$\hat{y}$.
			Then to show this field lines we have to define the x and y arrays of field foot points.\n
			``x0arr = linspace(0.0,10.0,20)``\n
			``y0arr = linspace(0.0,0.0,20)``\n
			``import pyPLUTO as pp``\n
			``D = pp.pload(45)``\n
			``I = pp.Image()``\n
			``I.myfieldlines(D,x0arr,y0arr,colors='k',ls='--',lw=1.0)``
		"""
				
		if len(x0arr) != len(y0arr) : print("Input Arrays should have same size")
		QxList=[]
		QyList=[]
		StreamFunction = []
		levels =[]
		if stream == True:
			X, Y = np.meshgrid(Data.x1,Data.x2.T)
			StreamFunction = X*(Data.Ax3.T)
			for i in range(len(x0arr)):
				nx = np.abs(X[0,:]-x0arr[i]).argmin()
				ny = np.abs(X[:,0]-y0arr[i]).argmin()
				levels.append(X[ny,nx]*Data.Ax3.T[ny,nx])
			contour(X,Y,StreamFunction,levels,colors=kwargs.get('colors'),linewidths=kwargs.get('lw',1),linestyles=kwargs.get('ls','solid'))
		else:
			for i in range(len(x0arr)):
				QxList.append(self.field_line(Data.bx1,Data.bx2,Data.x1,Data.x2,Data.dx1,Data.dx1,x0arr[i],y0arr[i]).get('qx'))
				QyList.append(self.field_line(Data.bx1,Data.bx2,Data.x1,Data.x2,Data.dx1,Data.dx1,x0arr[i],y0arr[i]).get('qy'))
				plot(QxList[i],QyList[i],color=kwargs.get('colors'))
				axis([min(Data.x1),max(Data.x1),min(Data.x2),max(Data.x2)])

	def getSphData(self,Data,w_dir=None,datatype=None,**kwargs):
		"""This method transforms the vector and scalar  fields from Spherical co-ordinates to Cylindrical.
		**Inputs**:
			Data -- pyPLUTO.pload object\n
			w_dir -- /path/to/the/working/directory/\n
			datatype -- If the data is of 'float' type then datatype = 'float' else by default the datatype is set to 'double'.
		*Optional Keywords*:
			rphi -- [Default] is set to False implies that the r-theta plane is transformed. If set True then the r-phi plane is transformed.\n
			x2cut -- Applicable for 3D data and it determines the co-ordinate of the x2 plane while r-phi is set to True.\n
			x3cut -- Applicable for 3D data and it determines the co-ordinate of the x3 plane while r-phi is set to False.
		"""
		Tool = Tools.Tools()
		key_value_pairs = []
		allvars = []
		for v in Data.vars:
			allvars.append(v)

		if kwargs.get('rphi',False)==True:
			R,TH = np.meshgrid(Data.x1,Data.x3)
			if Data.n3 != 1:
				for variable in allvars:
					key_value_pairs.append([variable,getattr(Data,variable)[:,kwargs.get('x2cut',0),:].T])

				SphData = dict(key_value_pairs)
				if ('bx1' in allvars) or ('bx2' in allvars):
					(SphData['b1c'],SphData['b3c']) = Tool.RTh2Cyl(R,TH,SphData.get('bx1'),SphData.get('bx3'))
					allvars.append('b1c')
					allvars.append('b3c')
				if ('vx1' in allvars) or ('vx2' in allvars):
					(SphData['v1c'],SphData['v3c']) = Tool.RTh2Cyl(R,TH,SphData.get('vx1'),SphData.get('vx3'))
					allvars.append('v1c')
					allvars.append('v3c')
			else:
				print("No x3 plane for 2D data")
		else:
			R,TH = np.meshgrid(Data.x1,Data.x2)
			if Data.n3 != 1:
				for variable in allvars:
					key_value_pairs.append([variable,getattr(Data,variable)[:,:,kwargs.get('x3cut',0)].T])
				SphData = dict(key_value_pairs)
				if ('bx1' in allvars) or ('bx2' in allvars):
					(SphData['b1c'],SphData['b2c']) = Tool.RTh2Cyl(R,TH,SphData.get('bx1'),SphData.get('bx2'))
					allvars.append('b1c')
					allvars.append('b2c')
				if ('vx1' in allvars) or ('vx2' in allvars):
					(SphData['v1c'],SphData['v2c']) = Tool.RTh2Cyl(R,TH,SphData.get('vx1'),SphData.get('vx2'))
					allvars.append('v1c')
					allvars.append('v2c')
			else:
				for variable in allvars:
					key_value_pairs.append([variable,getattr(Data,variable)[:,:].T])
				SphData = dict(key_value_pairs)
				if ('bx1' in allvars) or ('bx2' in allvars):
					(SphData['b1c'],SphData['b2c']) = Tool.RTh2Cyl(R,TH,SphData.get('bx1'),SphData.get('bx2'))
					allvars.append('b1c')
					allvars.append('b2c')
				if ('vx1' in allvars) or ('vx2' in allvars):
					(SphData['v1c'],SphData['v2c']) = Tool.RTh2Cyl(R,TH,SphData.get('vx1'),SphData.get('vx2'))
					allvars.append('v1c')
					allvars.append('v2c')
			
		for variable in allvars:
			if kwargs.get('rphi',False)==True:
				R,Z,SphData[variable]= Tool.sph2cyl(Data,SphData.get(variable),rphi=True,theta0=Data.x2[kwargs.get('x2cut',0)])
			else:
				if Data.n3 != 1:
					R,Z,SphData[variable] = Tool.sph2cyl(Data,SphData.get(variable),rphi=False)
				else:
					R,Z,SphData[variable] = Tool.sph2cyl(Data,SphData.get(variable),rphi=False)
		return R,Z,SphData
	 
	 
	def getPolarData(self, Data, ang_coord, rphi=False):
		"""To get the Cartesian Co-ordinates from Polar.
		**Inputs:**
			Data -- pyPLUTO pload Object\n
			ang_coord -- The Angular co-ordinate (theta or Phi)
		*Optional Keywords:*
			rphi -- Default value FALSE is for R-THETA data,
			Set TRUE for R-PHI data.\n
		**Outputs**:
			2D Arrays of X, Y from the Radius and Angular co-ordinates.\n
			They are used in pcolormesh in the Image.pldisplay functions.
		"""
		D = Data
		if ang_coord is D.x2:
			x2r = D.x2r
		elif ang_coord is D.x3:
			x2r = D.x3r
		else:
			print("Angular co-ordinate must be given")
				
		rcos = np.outer(np.cos(x2r), D.x1r)
		rsin = np.outer(np.sin(x2r), D.x1r)
		
		if rphi:
			xx = rcos
			yy = rsin
		else:
			xx = rsin
			yy = rcos

		return xx, yy

	def pltSphData(self,Data,w_dir=None,datatype=None,**kwargs):
		"""This method plots the transformed data obtained from getSphData using the matplotlib's imshow
		**Inputs:**
			Data -- pyPLUTO.pload object\n
			w_dir -- /path/to/the/working/directory/\n
			datatype -- Datatype.
		*Required Keywords*:
			plvar -- A string which represents the plot variable.\n
		*Optional Keywords*:
			logvar -- [Default = False] Set it True for plotting the log of a variable.\n
			rphi -- [Default = False - for plotting in r-theta plane] Set it True for plotting the variable in r-phi plane.
		"""
		if w_dir is None: w_dir=os.getcwd()+'/'
		R,Z,SphData = self.getSphData(Data,w_dir=w_dir,datatype=datatype,**kwargs)

		extent=(np.min(R.flat),max(R.flat),np.min(Z.flat),max(Z.flat))
		dRR=max(R.flat)-np.min(R.flat)
		dZZ=max(Z.flat)-np.min(Z.flat)

		isnotnan=~np.isnan(SphData[kwargs.get('plvar')])
		maxPl=max(SphData[kwargs.get('plvar')][isnotnan].flat)
		minPl=np.min(SphData[kwargs.get('plvar')][isnotnan].flat)
		normrange=False
		if minPl<0:
				normrange=True
		if maxPl>-minPl:
				minPl=-maxPl
		else:
				maxPl=-minPl
		if (normrange and kwargs.get('plvar')!='rho' and kwargs.get('plvar')!='prs'):
				SphData[kwargs.get('plvar')][-1][-1]=maxPl
				SphData[kwargs.get('plvar')][-1][-2]=minPl

		if (kwargs.get('logvar') == True):
				SphData[kwargs.get('plvar')] = np.log10(SphData[kwargs.get('plvar')])

		imshow(SphData[kwargs.get('plvar')], aspect='equal', origin='lower', cmap=cm.jet,extent=extent, interpolation='nearest')





from __future__ import division
import os
import sys
import array
import numpy as np
from . import Tools

####### Check for h5py to Read AMR data ######
try:
	import h5py as h5
	hasH5 = True
except ImportError:
	hasH5 = False
	
class pload(object):
	def __init__(self, ns, w_dir=None, datatype=None, level = 0, x1range=None, x2range=None, x3range=None):
		"""
		Loads the data.
		**Inputs**:
			ns -- Step Number of the data file\n
			w_dir -- path to the directory which has the data files\n
			datatype -- Datatype (default is set to read .dbl data files)
			level -- Integer value to represent AMR level (default = 0)
			x1range -- List with min and max value of x1 coordinates for zooming (applicable for AMR data)
			x2range -- List with min and max value of x2 coordinates for zooming (applicable for AMR data)
			x3range -- List with min and max value of x3 coordinates for zooming (applicable for AMR data)

		**Outputs**:
			pyPLUTO pload object whose keys are arrays of data values.
		"""
		self.NStep = ns
		self.Dt = 0.0

		self.n1 = 0
		self.n2 = 0
		self.n3 = 0

		self.x1 = []
		self.x2 = []
		self.x3 = []
		self.dx1 = []
		self.dx2 = []
		self.dx3 = []
		self.x1range = x1range
		self.x2range = x2range
		self.x3range = x3range
		self.NStepStr = str(self.NStep)
		while len(self.NStepStr) < 4:
			self.NStepStr = '0'+self.NStepStr

		if datatype is None:
			datatype = "dbl"
		self.datatype = datatype
		
		if ((not hasH5) and (datatype == 'hdf5')):
			print('To read AMR hdf5 files with python')
			print('Please install h5py (Python HDF5 Reader)')
			return
		self.level = level
		
		if w_dir is None:
			w_dir = os.getcwd() + '/'
		self.wdir = w_dir

		Data_dictionary = self.ReadDataFile(self.NStepStr)

		for keys in Data_dictionary:
			object.__setattr__(self, keys, Data_dictionary.get(keys))

	def ReadTimeInfo(self, timefile):
		""" 
		Read time info from the outfiles.
		**Inputs**:
			timefile -- name of the out file which has timing information.
		"""
		if (self.datatype == 'hdf5'):
			fh5 = h5.File(timefile,'r')
			self.SimTime = fh5.attrs.get('time')
			#self.Dt = 1.e-2 # Should be erased later given the level in AMR
			fh5.close()
		else:
			ns = self.NStep
			f_var = open(timefile, "r")
			tlist = []
			for line in f_var.readlines():
				tlist.append(line.split())
			self.SimTime = float(tlist[ns][1])
			self.Dt = float(tlist[ns][2])

	def ReadVarFile(self, varfile):
		""" 
		Read variable names from the outfiles.
		**Inputs**:
			varfile -- name of the out file which has variable information.
		"""
		if (self.datatype == 'hdf5'):
			fh5 = h5.File(varfile,'r')
			self.filetype = 'single_file'
			self.endianess = '>' # not used with AMR, kept for consistency
			self.vars = []
			for iv in range(fh5.attrs.get('num_components')):
					self.vars.append(fh5.attrs.get('component_'+str(iv)))
			fh5.close()
		else:
			vfp = open(varfile, "r")
			varinfo = vfp.readline().split()
			self.filetype = varinfo[4]
			self.endianess = varinfo[5]
			self.vars = varinfo[6:]
			vfp.close()

	def ReadGridFile(self, gridfile):
		""" Read grid values from the grid.out file.
		**Inputs**:
			gridfile -- name of the grid.out file which has information about the grid.
		"""
		xL = []
		xR = []
		nmax = []
		gfp = open(gridfile, "r")
		for i in gfp.readlines():
			if len(i.split()) == 1:
				try:
					int(i.split()[0])
					nmax.append(int(i.split()[0]))
				except:
					pass
					
			if len(i.split()) == 3:
				try:
					int(i.split()[0])
					xL.append(float(i.split()[1]))
					xR.append(float(i.split()[2]))
				except:
					if (i.split()[1] == 'GEOMETRY:'):
						self.geometry=i.split()[2]
					pass
							
		self.n1, self.n2, self.n3 = nmax
		n1 = self.n1
		n1p2 = self.n1 + self.n2
		n1p2p3 = self.n1 + self.n2 + self.n3
		self.x1 = np.asarray([0.5*(xL[i]+xR[i]) for i in range(n1)])
		self.dx1 = np.asarray([(xR[i]-xL[i]) for i in range(n1)])
		self.x2 = np.asarray([0.5*(xL[i]+xR[i]) for i in range(n1, n1p2)])
		self.dx2 = np.asarray([(xR[i]-xL[i]) for i in range(n1, n1p2)])
		self.x3 = np.asarray([0.5*(xL[i]+xR[i]) for i in range(n1p2, n1p2p3)])
		self.dx3 = np.asarray([(xR[i]-xL[i]) for i in range(n1p2, n1p2p3)])


		# Stores the total number of points in '_tot' variable in case only
		# a portion of the domain is loaded. Redefine the x and dx arrays
		# to match the requested ranges
		self.n1_tot = self.n1 ; self.n2_tot = self.n2 ; self.n3_tot = self.n3
		if (self.x1range != None):
			self.n1_tot = self.n1
			self.irange = range(abs(self.x1-self.x1range[0]).argmin(),abs(self.x1-self.x1range[1]).argmin()+1)
			self.n1  = len(self.irange)
			self.x1  = self.x1[self.irange]
			self.dx1 = self.dx1[self.irange]
		else:
			self.irange = range(self.n1)
		if (self.x2range != None):
			self.n2_tot = self.n2
			self.jrange = range(abs(self.x2-self.x2range[0]).argmin(),abs(self.x2-self.x2range[1]).argmin()+1)
			self.n2  = len(self.jrange)
			self.x2  = self.x2[self.jrange]
			self.dx2 = self.dx2[self.jrange]
		else:
			self.jrange = range(self.n2)
		if (self.x3range != None):
			self.n3_tot = self.n3
			self.krange = range(abs(self.x3-self.x3range[0]).argmin(),abs(self.x3-self.x3range[1]).argmin()+1)
			self.n3  = len(self.krange)
			self.x3  = self.x3[self.krange]
			self.dx3 = self.dx3[self.krange]
		else:
			self.krange = range(self.n3)
		self.Slice=(self.x1range != None) or (self.x2range != None) or (self.x3range != None)


		# Create the xr arrays containing the edges positions
		# Useful for pcolormesh which should use those
		self.x1r = np.zeros(len(self.x1)+1) ; self.x1r[1:] = self.x1 + self.dx1/2.0 ; self.x1r[0] = self.x1r[1]-self.dx1[0]
		self.x2r = np.zeros(len(self.x2)+1) ; self.x2r[1:] = self.x2 + self.dx2/2.0 ; self.x2r[0] = self.x2r[1]-self.dx2[0]
		self.x3r = np.zeros(len(self.x3)+1) ; self.x3r[1:] = self.x3 + self.dx3/2.0 ; self.x3r[0] = self.x3r[1]-self.dx3[0]


		prodn = self.n1*self.n2*self.n3
		if prodn == self.n1:
			self.nshp = (self.n1)
		elif prodn == self.n1*self.n2:
			self.nshp = (self.n2, self.n1)
		else:
			self.nshp = (self.n3, self.n2, self.n1)


	def DataScanVTK(self, fp, n1, n2, n3, endian, dtype):
		""" Scans the VTK data files.

		**Inputs**:
		fp -- Data file pointer\n
		n1 -- No. of points in X1 direction\n
		n2 -- No. of points in X2 direction\n
		n3 -- No. of points in X3 direction\n
		endian -- Endianess of the data\n
		dtype -- datatype
		**Output**:
			Dictionary consisting of variable names as keys and its values.
		"""
		ks = []
		vtkvar = []
		while True:
			line = fp.readline()
			try:
				line.split()[0]
			except IndexError as error:
				pass
			else:
				if line.split()[0] == b'SCALARS':
					ks.append(line.decode().split()[1])
				elif line.split()[0] == b'LOOKUP_TABLE':
					A = array.array(dtype)
					fmt = endian+str(n1*n2*n3)+dtype
					nb = np.dtype(fmt).itemsize
					A.frombytes(fp.read(nb))
					if (self.Slice):
						darr = np.zeros((n1*n2*n3))
						indxx = np.sort([self.n3_tot*self.n2_tot*k + j*self.n2_tot + i for i in self.irange for j in self.jrange for k in self.krange])
						if (sys.byteorder != self.endianess):
							A.byteswap()
						for ii,iii in enumerate(indxx):
							darr[ii] = A[iii]
						vtkvar_buf = [darr]
					else:
						vtkvar_buf = np.frombuffer(A,dtype=np.dtype(fmt))
					vtkvar.append(np.reshape(vtkvar_buf,self.nshp).transpose())
				else:
					pass
			if line == b'':
				break

		vtkvardict = dict(zip(ks,vtkvar))
		return vtkvardict
					
	def DataScanHDF5(self, fp, myvars, ilev):
		""" Scans the Chombo HDF5 data files for AMR in PLUTO.
		**Inputs**:
			fp     -- Data file pointer\n
			myvars -- Names of the variables to read\n
			ilev   -- required AMR level
		**Output**:
			Dictionary consisting of variable names as keys and its values.
		**Note**:
			Due to the particularity of AMR, the grid arrays loaded in ReadGridFile are overwritten here.
		"""
		# Read the grid information
		dim = fp['Chombo_global'].attrs.get('SpaceDim')
		nlev = fp.attrs.get('num_levels')
		il = min(nlev-1,ilev)
		lev  = []
		for i in range(nlev):
			lev.append('level_'+str(i))
		freb = np.zeros(nlev,dtype='int')
		for i in range(il+1)[::-1]:
			fl = fp[lev[i]]
			if (i == il):
				pdom = fl.attrs.get('prob_domain')
				dx = fl.attrs.get('dx')
				dt = fl.attrs.get('dt')
				ystr = 1. ; zstr = 1. ; logr = 0
				try:
					geom = fl.attrs.get('geometry')
					logr = fl.attrs.get('logr')
					if (dim == 2):ystr = fl.attrs.get('g_x2stretch')
					elif (dim == 3):zstr = fl.attrs.get('g_x3stretch')
				except:
					print('Old HDF5 file, not reading stretch and logr factors')
				freb[i] = 1
				x1b = fl.attrs.get('domBeg1')
				if (dim == 1):x2b = 0
				else:x2b = fl.attrs.get('domBeg2')
				if (dim == 1 or dim == 2):x3b = 0
				else:x3b = fl.attrs.get('domBeg3')
				jbeg = 0 ; jend = 0 ; ny = 1
				kbeg = 0 ; kend = 0 ; nz = 1
				if (dim == 1):
					ibeg = pdom[0] ; iend = pdom[1] ; nx = iend-ibeg+1
				elif (dim == 2):
					ibeg = pdom[0] ; iend = pdom[2] ; nx = iend-ibeg+1
					jbeg = pdom[1] ; jend = pdom[3] ; ny = jend-jbeg+1
				elif (dim == 3):
					ibeg = pdom[0] ; iend = pdom[3] ; nx = iend-ibeg+1
					jbeg = pdom[1] ; jend = pdom[4] ; ny = jend-jbeg+1
					kbeg = pdom[2] ; kend = pdom[5] ; nz = kend-kbeg+1
			else:
				rat = fl.attrs.get('ref_ratio')
				freb[i] = rat*freb[i+1]
	
		dx0 = dx*freb[0]

		## Allow to load only a portion of the domain
		if (self.x1range != None):
			if logr == 0:self.x1range = self.x1range-x1b
			else:self.x1range = [np.log(self.x1range[0]/x1b),np.log(self.x1range[1]/x1b)]
			ibeg0 = min(self.x1range)/dx0 ; iend0 = max(self.x1range)/dx0
			ibeg  = max([ibeg, int(ibeg0*freb[0])]) ; iend = min([iend,int(iend0*freb[0]-1)])
			nx = iend-ibeg+1
		if (self.x2range != None):
			self.x2range = (self.x2range-x2b)/ystr
			jbeg0 = min(self.x2range)/dx0 ; jend0 = max(self.x2range)/dx0
			jbeg  = max([jbeg, int(jbeg0*freb[0])]) ; jend = min([jend,int(jend0*freb[0]-1)])
			ny = jend-jbeg+1
		if (self.x3range != None):
			self.x3range = (self.x3range-x3b)/zstr
			kbeg0 = min(self.x3range)/dx0 ; kend0 = max(self.x3range)/dx0
			kbeg  = max([kbeg, int(kbeg0*freb[0])]) ; kend = min([kend,int(kend0*freb[0]-1)])
			nz = kend-kbeg+1
		
		## Create uniform grids at the required level
		if logr == 0:
			x1 = x1b + (ibeg+np.array(range(nx))+0.5)*dx
		else:
			x1 = x1b*(np.exp((ibeg+np.array(range(nx))+1)*dx)+np.exp((ibeg+np.array(range(nx)))*dx))*0.5
		x2 = x2b + (jbeg+np.array(range(ny))+0.5)*dx*ystr
		x3 = x3b + (kbeg+np.array(range(nz))+0.5)*dx*zstr
		if logr == 0:
			dx1 = np.ones(nx)*dx
		else:
			dx1 = x1b*(np.exp((ibeg+np.array(range(nx))+1)*dx)-np.exp((ibeg+np.array(range(nx)))*dx))
		dx2 = np.ones(ny)*dx*ystr
		dx3 = np.ones(nz)*dx*zstr

		# Create the xr arrays containing the edges positions
		# Useful for pcolormesh which should use those
		x1r = np.zeros(len(x1)+1) ; x1r[1:] = x1 + dx1/2.0 ; x1r[0] = x1r[1]-dx1[0]
		x2r = np.zeros(len(x2)+1) ; x2r[1:] = x2 + dx2/2.0 ; x2r[0] = x2r[1]-dx2[0]
		x3r = np.zeros(len(x3)+1) ; x3r[1:] = x3 + dx3/2.0 ; x3r[0] = x3r[1]-dx3[0]
		NewGridDict = dict([('n1',nx),('n2',ny),('n3',nz),\
							('x1',x1),('x2',x2),('x3',x3),\
							('x1r',x1r),('x2r',x2r),('x3r',x3r),\
							('dx1',dx1),('dx2',dx2),('dx3',dx3),\
							('Dt',dt)])
		
		# Variables table
		nvar = len(myvars)
		vars = np.zeros((nx,ny,nz,nvar))
		LevelDic = {'nbox':0,'ibeg':ibeg,'iend':iend,'jbeg':jbeg,'jend':jend,'kbeg':kbeg,'kend':kend}
		AMRLevel = []
		AMRBoxes = np.zeros((nx,ny,nz))
		for i in range(il+1):
			AMRLevel.append(LevelDic.copy())
			fl = fp[lev[i]]
			data = fl['data:datatype=0']
			boxes = fl['boxes']
			nbox = len(boxes['lo_i'])
			AMRLevel[i]['nbox'] = nbox
			ncount = int(0)
			AMRLevel[i]['box']=[]
			for j in range(nbox): # loop on all boxes of a given level
				AMRLevel[i]['box'].append({'x0':0.,'x1':0.,'ib':int(0),'ie':int(0),\
											'y0':0.,'y1':0.,'jb':int(0),'je':int(0),\
											'z0':0.,'z1':0.,'kb':int(0),'ke':int(0)})
				# Box indexes
				ib = boxes[j]['lo_i'] ; ie = boxes[j]['hi_i'] ; nbx = ie-ib+1
				jb = 0 ; je = 0 ; nby = 1
				kb = 0 ; ke = 0 ; nbz = 1
				if (dim > 1):
					jb = boxes[j]['lo_j'] ; je = boxes[j]['hi_j'] ; nby = je-jb+1
				if (dim > 2):
					kb = boxes[j]['lo_k'] ; ke = boxes[j]['hi_k'] ; nbz = ke-kb+1
				szb = nbx*nby*nbz*nvar
				# Rescale to current level
				kb = kb*freb[i] ; ke = (ke+1)*freb[i] - 1
				jb = jb*freb[i] ; je = (je+1)*freb[i] - 1
				ib = ib*freb[i] ; ie = (ie+1)*freb[i] - 1
	
				# Skip boxes lying outside ranges
				if ((ib > iend) or (ie < ibeg) or \
					(jb > jend) or (je < jbeg) or \
					(kb > kend) or (ke < kbeg)):
					ncount = ncount + szb
				else:
					### Read data
					q = data[ncount:ncount+szb].reshape((nvar,nbz,nby,nbx)).T
				
					### Find boxes intersections with current domain ranges
					ib0 = max([ibeg,ib]) ; ie0 = min([iend,ie])
					jb0 = max([jbeg,jb]) ; je0 = min([jend,je])
					kb0 = max([kbeg,kb]) ; ke0 = min([kend,ke])
		
					### Store box corners in the AMRLevel structure
					if logr == 0:
						AMRLevel[i]['box'][j]['x0'] = x1b + dx*(ib0)
						AMRLevel[i]['box'][j]['x1'] = x1b + dx*(ie0+1)
					else:
						AMRLevel[i]['box'][j]['x0'] = x1b*np.exp(dx*(ib0))
						AMRLevel[i]['box'][j]['x1'] = x1b*np.exp(dx*(ie0+1))
					AMRLevel[i]['box'][j]['y0'] = x2b + dx*(jb0)*ystr
					AMRLevel[i]['box'][j]['y1'] = x2b + dx*(je0+1)*ystr
					AMRLevel[i]['box'][j]['z0'] = x3b + dx*(kb0)*zstr
					AMRLevel[i]['box'][j]['z1'] = x3b + dx*(ke0+1)*zstr
					AMRLevel[i]['box'][j]['ib'] = ib0 ; AMRLevel[i]['box'][j]['ie'] = ie0
					AMRLevel[i]['box'][j]['jb'] = jb0 ; AMRLevel[i]['box'][j]['je'] = je0
					AMRLevel[i]['box'][j]['kb'] = kb0 ; AMRLevel[i]['box'][j]['ke'] = ke0
					AMRBoxes[ib0-ibeg:ie0-ibeg+1, jb0-jbeg:je0-jbeg+1, kb0-kbeg:ke0-kbeg+1] = il
					
					### Extract the box intersection from data stored in q
					cib0 = (ib0-ib)//freb[i] ; cie0 = (ie0-ib)//freb[i]
					cjb0 = (jb0-jb)//freb[i] ; cje0 = (je0-jb)//freb[i]
					ckb0 = (kb0-kb)//freb[i] ; cke0 = (ke0-kb)//freb[i]
					q1 = np.zeros((cie0-cib0+1, cje0-cjb0+1, cke0-ckb0+1,nvar))
					q1 = q[cib0:cie0+1,cjb0:cje0+1,ckb0:cke0+1,:]
					
					# Remap the extracted portion
					if (dim == 1):
						new_shape = (ie0-ib0+1,1)
					elif (dim == 2):
						new_shape = (ie0-ib0+1,je0-jb0+1)
					else:
						new_shape = (ie0-ib0+1,je0-jb0+1,ke0-kb0+1)
							
					stmp = list(new_shape)
					while stmp.count(1) > 0:
						stmp.remove(1)
					new_shape = tuple(stmp)
					
					myT = Tools.Tools()
					for iv in range(nvar):
						vars[ib0-ibeg:ie0-ibeg+1,jb0-jbeg:je0-jbeg+1,kb0-kbeg:ke0-kbeg+1,iv] = \
							myT.congrid(q1[:,:,:,iv].squeeze(),new_shape,method='linear',minusone=True).reshape((ie0-ib0+1,je0-jb0+1,ke0-kb0+1))
					ncount = ncount+szb

		h5vardict={}
		for iv in range(nvar):
			myvars[iv] = myvars[iv].decode()
			h5vardict[myvars[iv]] = vars[:,:,:,iv].squeeze()
		AMRdict = dict([('AMRBoxes',AMRBoxes),('AMRLevel',AMRLevel)])
		OutDict = dict(NewGridDict)
		OutDict.update(AMRdict)
		OutDict.update(h5vardict)
		return OutDict

	
	def DataScan(self, fp, n1, n2, n3, endian, dtype, off=None):
		""" Scans the data files in all formats.
		**Inputs**:
			fp -- Data file pointer\n
			n1 -- No. of points in X1 direction\n
			n2 -- No. of points in X2 direction\n
			n3 -- No. of points in X3 direction\n
			endian -- Endianess of the data\n
			dtype -- datatype, eg : dbl, flt, vtk, hdf5\n
			off -- offset (for avoiding staggered B fields)
		**Output**:
			Dictionary consisting of variable names as keys and its values.
		"""
		if off is not None:
			off_fmt = endian+str(off)+dtype
			nboff = np.dtype(off_fmt).itemsize
			fp.read(nboff)

		n1_tot = self.n1_tot ; n2_tot = self.n2_tot; n3_tot = self.n3_tot
		A = array.array(dtype)
		fmt = endian+str(n1_tot*n2_tot*n3_tot)+dtype
		nb = np.dtype(fmt).itemsize
		A.frombytes(fp.read(nb))
		
		if (self.Slice):
			darr = np.zeros((n1*n2*n3))
			indxx = np.sort([n3_tot*n2_tot*k + j*n2_tot + i for i in self.irange for j in self.jrange for k in self.krange])
			if (sys.byteorder != self.endianess):
					A.byteswap()
			for ii,iii in enumerate(indxx):
					darr[ii] = A[iii]
			darr = [darr]
		else:
			darr = np.frombuffer(A,dtype=np.dtype(fmt))
		
		return np.reshape(darr[0],self.nshp).transpose()

	def ReadSingleFile(self, datafilename, myvars, n1, n2, n3, endian,
										 dtype, ddict):
		"""Reads a single data file, data.****.dtype.
		**Inputs**:
			datafilename -- Data file name\n
		myvars -- List of variable names to be read\n
			n1 -- No. of points in X1 direction\n
			n2 -- No. of points in X2 direction\n
			n3 -- No. of points in X3 direction\n
			endian -- Endianess of the data\n
			dtype -- datatype\n
			ddict -- Dictionary containing Grid and Time Information
			which is updated
		**Output**:
			Updated Dictionary consisting of variable names as keys and its values.
		"""
		if self.datatype == 'hdf5':
			fp = h5.File(datafilename,'r')
		else:
			fp = open(datafilename, "rb")
		print("Reading single data file : %s"%datafilename)
		
		if self.datatype == 'vtk':
			vtkd = self.DataScanVTK(fp, n1, n2, n3, endian, dtype)
			ddict.update(vtkd)
		elif self.datatype == 'hdf5':
			h5d = self.DataScanHDF5(fp,myvars,self.level)
			ddict.update(h5d)
		else:
			for i in range(len(myvars)):
				if myvars[i] == 'bx1s':
					ddict.update({myvars[i]: self.DataScan(fp, n1, n2, n3, endian,dtype, off=n2*n3)})
				elif myvars[i] == 'bx2s':
					ddict.update({myvars[i]: self.DataScan(fp, n1, n2, n3, endian,dtype, off=n3*n1)})
				elif myvars[i] == 'bx3s':
					ddict.update({myvars[i]: self.DataScan(fp, n1, n2, n3, endian,dtype, off=n1*n2)})
				else:
					ddict.update({myvars[i]: self.DataScan(fp, n1, n2, n3, endian, dtype)})
		fp.close()

	def ReadMultipleFiles(self, nstr, dataext, myvars, n1, n2, n3, endian,
												dtype, ddict):
		"""Reads a  multiple data files, varname.****.dataext.
		**Inputs**:
			nstr -- File number in form of a string\n
			dataext -- Data type of the file, e.g., 'dbl', 'flt' or 'vtk' \n
			myvars -- List of variable names to be read\n
			n1 -- No. of points in X1 direction\n
			n2 -- No. of points in X2 direction\n
			n3 -- No. of points in X3 direction\n
			endian -- Endianess of the data\n
			dtype -- datatype\n
			ddict -- Dictionary containing Grid and Time Information
			which is updated.
		**Output**:
			Updated Dictionary consisting of variable names as keys and its values.
		"""
		for i in range(len(myvars)):
			datafilename = self.wdir+myvars[i]+"."+nstr+dataext
			print("Reading multiple variable files : %s"%datafilename)
			fp = open(datafilename, "rb")
			if self.datatype == 'vtk':
				ddict.update(self.DataScanVTK(fp, n1, n2, n3, endian, dtype))
			else:
				ddict.update({myvars[i]: self.DataScan(fp, n1, n2, n3, endian,dtype)})
			fp.close()

	def ReadDataFile(self, num):
		"""Reads the data file generated from PLUTO code.
		**Inputs**:
			num -- Data file number in form of an Integer.
		**Outputs**:
			Dictionary that contains all information about Grid, Time and
			variables.
		"""
		gridfile = self.wdir+"grid.out"
		if self.datatype == "flt":
			dtype = "f"
			varfile = self.wdir+"flt.out"
			dataext = ".flt"
		elif self.datatype == "vtk":
			dtype = "f"
			varfile = self.wdir+"vtk.out"
			dataext=".vtk"
		elif self.datatype == 'hdf5':
			dtype = 'd'
			dataext = '.hdf5'
			nstr = num
			varfile = self.wdir+"data."+nstr+dataext
		else:
			dtype = "d"
			varfile = self.wdir+"dbl.out"
			dataext = ".dbl"

		self.ReadVarFile(varfile)
		self.ReadGridFile(gridfile)
		self.ReadTimeInfo(varfile)
		nstr = num
		if self.endianess == 'big':
			endian = ">"
		elif self.datatype == 'vtk':
			endian = ">"
		else:
			endian = "<"

		D = [('NStep', self.NStep), ('SimTime', self.SimTime), ('Dt', self.Dt),
			('n1', self.n1), ('n2', self.n2), ('n3', self.n3),
			('x1', self.x1), ('x2', self.x2), ('x3', self.x3),
			('dx1', self.dx1), ('dx2', self.dx2), ('dx3', self.dx3),
			('endianess', self.endianess), ('datatype', self.datatype),
			('filetype', self.filetype)]
		ddict = dict(D)

		if self.filetype == "single_file":
			datafilename = self.wdir+"data."+nstr+dataext
			self.ReadSingleFile(datafilename, self.vars, self.n1, self.n2,
								self.n3, endian, dtype, ddict)
		elif self.filetype == "multiple_files":
			self.ReadMultipleFiles(nstr, dataext, self.vars, self.n1, self.n2,
									self.n3, endian, dtype, ddict)
		else:
			print("Wrong file type : CHECK pluto.ini for file type.")
			print("Only supported are .dbl, .flt, .vtk, .hdf5")
			sys.exit()

		return ddict
